import traceback

import numpy as np

from ..handler.nativeMethods import NativeMethods
from ..helper.decorators import Singleton, deprecated
from ..helper.messages import Message
from ..helper.result_objects.compareResult import FaceCompareResult
from ..helper.result_objects.deleteResult import FaceDeleteResult
from ..helper.result_objects.enrollPredictResult import FaceEnrollPredictResult
from ..helper.result_objects.faceValidationResult import FaceValidationResult
from ..helper.result_objects.isValidDeprecatedResult import FaceIsValidDeprecatedResult
from ..helper.result_objects.isoFaceResult import ISOFaceResult
from ..helper.utils import FaceValidationCode
from ..settings.cacheType import CacheType
from ..settings.configuration import ConfigObject
from ..settings.loggingLevel import LoggingLevel


class Face(metaclass=Singleton):

    def __init__(self, api_key: str, server_url: str, local_storage_path: str, logging_level: LoggingLevel,
                 tf_num_thread: int, cache_type: CacheType, config_object: ConfigObject = None):
        self.message = Message()
        self.face_factor_processor = NativeMethods(api_key=api_key, server_url=server_url,
                                                   local_storage_path=local_storage_path, logging_level=logging_level,
                                                   tf_num_thread=tf_num_thread, cache_type=cache_type,
                                                   config_object=config_object)

    def update_config(self, config_object):
        self.face_factor_processor.update_config(config_object=config_object)

    def enroll(self, image_data: np.array, config_object: ConfigObject = None) -> FaceEnrollPredictResult:
        try:
            json_data = self.face_factor_processor.enroll(image_data, config_object=config_object)
            if not json_data:
                return FaceEnrollPredictResult(message=self.message.EXCEPTION_ERROR_ENROLL)
            if "PI" not in json_data:
                return FaceEnrollPredictResult(status=json_data.get("status", -1),
                                               message=json_data.get("message", self.message.EXCEPTION_ERROR_ENROLL))
            return FaceEnrollPredictResult(enroll_level=json_data["PI"].get("enroll_level", None),
                                           uuid=json_data["PI"].get("uuid", None),
                                           guid=json_data["PI"].get("guid", None),
                                           token=json_data["PI"].get("token", None), status=json_data.get("status", -1),
                                           message=json_data.get("message", ""))
        except Exception as e:
            print(e, traceback.format_exc())
            return FaceEnrollPredictResult(message=self.message.EXCEPTION_ERROR_ENROLL)

    def predict(self, image_data: np.array, config_object: ConfigObject = None) -> FaceEnrollPredictResult:
        try:
            json_data = self.face_factor_processor.predict(image_data, config_object=config_object)
            if not json_data:
                return FaceEnrollPredictResult(message=self.message.EXCEPTION_ERROR_PREDICT)
            if "PI" not in json_data:
                return FaceEnrollPredictResult(status=json_data.get("status", -1),
                                               message=json_data.get("message", self.message.EXCEPTION_ERROR_PREDICT))
            return FaceEnrollPredictResult(enroll_level=json_data["PI"].get("enroll_level", None),
                                           uuid=json_data["PI"].get("uuid", None),
                                           guid=json_data["PI"].get("guid", None),
                                           token=json_data["PI"].get("token", None), status=json_data.get("status", -1),
                                           message=json_data.get("message", ""))
        except Exception as e:
            print(e, traceback.format_exc())
            return FaceEnrollPredictResult(message=self.message.EXCEPTION_ERROR_PREDICT)

    def delete(self, uuid: str) -> FaceDeleteResult:
        try:
            json_response = self.face_factor_processor.delete(uuid)
            if not json_response:
                return FaceDeleteResult(message=self.message.EXCEPTION_ERROR_DELETE)
            return FaceDeleteResult(status=json_response.get("status", -1), message=json_response.get("message", ""))
        except Exception as e:
            print(e, traceback.format_exc())
            return FaceDeleteResult(message=self.message.EXCEPTION_ERROR_DELETE)

    def compare(self, image_data_1: np.array, image_data_2: np.array,
                config_object: ConfigObject = None) -> FaceCompareResult:
        try:
            json_data = self.face_factor_processor.compare_files(image_data_1, image_data_2,
                                                                 config_object=config_object)
            if not json_data:
                return FaceCompareResult(message=self.message.EXCEPTION_ERROR_COMPARE)

            return FaceCompareResult(result=json_data.get("result", None),
                                     distance_min=json_data.get("distance_min", None),
                                     first_validation_result=json_data.get("valid_flag_a", None),
                                     second_validation_result=json_data.get("valid_flag_b", None),
                                     distance_max=json_data.get("distance_max", None),
                                     distance_mean=json_data.get("distance_mean", None),
                                     status=json_data.get("status", 1), message=json_data.get("message", ""))
        except Exception as e:
            print(e, traceback.format_exc())
            return FaceCompareResult(message=self.message.EXCEPTION_ERROR_COMPARE)

    @deprecated
    def is_valid_deprecated(self, image_data: np.array,
                            config_object: ConfigObject = None) -> FaceIsValidDeprecatedResult:
        try:
            json_data = self.face_factor_processor.is_valid(image_data, config_object=config_object)
            if not json_data:
                return FaceIsValidDeprecatedResult(message=self.message.IS_VALID_ERROR)

            result_ = None
            if json_data.get("result", None) in FaceValidationCode:
                result_ = FaceValidationCode(json_data.get("result", None)).name

            return FaceIsValidDeprecatedResult(result=result_, age_factor=json_data.get("ageFactor", None), status=0)

        except Exception as e:
            print(e, traceback.format_exc())
            return FaceIsValidDeprecatedResult(message=self.message.IS_VALID_ERROR)

    def is_valid(self, image_data: np.array, config_object: ConfigObject = None) -> FaceValidationResult:
        try:
            json_data = self.face_factor_processor.is_valid_without_age(image_data, config_object=config_object)
            if not json_data:
                return FaceValidationResult(message=self.message.IS_VALID_ERROR)

            if json_data.get("error", -100) != 0:
                return FaceValidationResult(error=json_data.get("error", -100), message=self.message.IS_VALID_ERROR)

            face_validate_result_object = FaceValidationResult(error=json_data.get("error", -1), message="OK")
            for face in json_data.get("faces", []):
                _return_code = face.get("status", -100)
                if _return_code in FaceValidationCode:
                    _message = FaceValidationCode(_return_code).name
                else:
                    raise Exception("Status code out of bounds.")
                face_validate_result_object.append_face_objects(return_code=_return_code, message=_message,
                                                                top_left_coordinate=face["box"].get("top_left", None),
                                                                bottom_right_coordinate=face["box"].get("bottom_right",
                                                                                                        None))

            return face_validate_result_object
        except Exception as e:
            print(e, traceback.format_exc())
            return FaceValidationResult(message=self.message.IS_VALID_ERROR)

    def estimate_age(self, image_data: np.array, config_object: ConfigObject = None) -> FaceValidationResult:
        try:
            json_data = self.face_factor_processor.estimate_age(image_data, config_object=config_object)
            if not json_data:
                return FaceValidationResult(message=self.message.AGE_ESTIMATE_ERROR)

            if json_data.get("error", -1) != 0:
                return FaceValidationResult(error=json_data.get("error", -1), message=self.message.AGE_ESTIMATE_ERROR)

            face_age_result_object = FaceValidationResult(error=json_data.get("error", -1), message="OK")
            for face in json_data.get("faces"):

                _return_code = face.get("status", -1)
                _age = face.get("age", -1.0)
                if _return_code in FaceValidationCode:
                    _message = FaceValidationCode(_return_code).name
                else:
                    raise Exception("Status code out of bounds.")
                if _return_code == -1:
                    _age = -1
                face_age_result_object.append_face_objects(return_code=_return_code, age=_age, message=_message,
                                                           top_left_coordinate=face["box"].get("top_left", None),
                                                           bottom_right_coordinate=face["box"].get("bottom_right",
                                                                                                   None))

            return face_age_result_object
        except Exception as e:
            print(e, traceback.format_exc())
            return FaceValidationResult(message=self.message.AGE_ESTIMATE_ERROR)

    def get_iso_face(self, image_data: np.array, config_object: ConfigObject = None) -> ISOFaceResult:
        try:
            json_data = self.face_factor_processor.get_iso_face(image_data, config_object=config_object)
            if not json_data:
                return ISOFaceResult(message=self.message.EXCEPTION_ERROR_GET_ISO_FACE)

            if json_data.get("status", -1) != 0:
                return ISOFaceResult(status=json_data.get("status", -1),
                                     message=self.message.EXCEPTION_ERROR_GET_ISO_FACE)

            return ISOFaceResult(iso_image_width=json_data.get("iso_image_width", None),
                                 iso_image_height=json_data.get("iso_image_height", None),
                                 iso_image_channels=json_data.get("iso_image_channels", None),
                                 confidence=json_data.get("confidence", None),
                                 image=json_data.get("image", None),
                                 status=json_data.get("status", -1), message=json_data.get("message", "OK"))

        except Exception as e:
            print(e, traceback.format_exc())
            return ISOFaceResult(message=self.message.EXCEPTION_ERROR_GET_ISO_FACE)



class FaceAgeResult:
    def __init__(self, age=None, status=-1, message=""):
        self._status = status
        self._message = message
        self._age = age

    @property
    def status(self):
        """
        Returns the status of the operation

        0 - If successfully obtained result from server

        -1 - In case of error

        """
        return self._status

    @property
    def message(self):
        """
        Returns the message of the operation
        """
        return self._message

    @property
    def age(self):
        """
        Returns the predicted age of the image

        Age value might be in the float format as it is returned from DNN model
        """
        return self._age

    @status.setter
    def status(self, value):
        self._status = value

    @message.setter
    def message(self, value):
        self._message = value

    @age.setter
    def age(self, value):
        self._age = value



from enum import Enum


class LoggingLevel(Enum):
    """
    LoggingLevel on how much debug data has to be displayed

    off - No logs

    minimal - Important logs on operation

    normal - Default log

    full - Complete logs

    """
    off = 0
    minimal = 1
    normal = 2
    full = 3



from ..utils import BoundingBox


class FaceObjectResult(object):

    def __init__(self, return_code=-100, message="", age=None, top_left_coordinate=None, bottom_right_coordinate=None):
        self._return_code = return_code
        self._message = message
        self._age = age
        self._bounding_box = BoundingBox(top_left_coordinate, bottom_right_coordinate)

    @property
    def return_code(self) -> int:
        """
        Returns the return code for the image
        """
        return self._return_code

    @property
    def message(self) -> str:
        """
        Returns the message of the operation
        """
        return self._message

    @property
    def age(self) -> float:
        """
        Returns the predicted age of the image

        Age value might be in the float format as it is returned from DNN model
        """
        return self._age

    @property
    def bounding_box(self) -> BoundingBox:
        """Returns the bounding box for the images found

        Returns
        -------
        BoundingBox
            top_left_coordinate: Point

            bottom_right_coordinate: Point
        """
        return self._bounding_box

    @return_code.setter
    def return_code(self, value):
        self._return_code = value

    @message.setter
    def message(self, value):
        self._message = value

    @age.setter
    def age(self, value):
        self._age = value

    @bounding_box.setter
    def bounding_box(self, value):
        self._bounding_box = value



import ctypes
import json
import pathlib
import sys
from ctypes import *
from typing import Any

import numpy as np
from PIL import Image

from ..settings.cacheType import CacheType
from ..settings.configuration import ConfigObject
from ..settings.loggingLevel import LoggingLevel


class NativeMethods(object):
    def __init__(self, api_key: str, server_url: str, local_storage_path: str,
                 logging_level: LoggingLevel, tf_num_thread: int, cache_type: CacheType,
                 config_object: ConfigObject = None):
        try:
            self._config_object = config_object
            self._library_path = str(pathlib.Path(__file__).parent.joinpath("lib/lib_fhe.so").resolve())
            self._embedding_length = 128
            self._num_embeddings = 80
            self._aug_size = 224 * 224 * 4 * self._num_embeddings
            self._spl_so_face = None
            self._tf_num_thread = tf_num_thread
            self._api_key = bytes(api_key, 'utf-8')
            self._server_url = bytes(server_url, 'utf-8')
            self._logging_level = logging_level
            self._local_storage_path = local_storage_path
            self._cache_type = cache_type
            self._face_setup()
        except Exception as e:
            print("Error ", e)
            sys.exit(1)

    def update_config(self, config_object):
        self._config_object = config_object
        if self._config_object and self._config_object.get_config_param():
            c_config_param = c_char_p(bytes(self._config_object.get_config_param(), 'utf-8'))
            c_config_param_len = c_int(len(self._config_object.get_config_param()))
            self._spl_so_face.privid_set_configuration(self._spl_so_face.new_handle, c_config_param, c_config_param_len)

    def _face_setup(self):
        self._spl_so_face = ctypes.CDLL(self._library_path)

        # privid_global_settings
        self._spl_so_face.privid_global_settings.argtypes = [c_uint8, c_uint8]
        self._spl_so_face.privid_global_settings.restype = c_bool
        self._spl_so_face.privid_global_settings(self._tf_num_thread, self._logging_level.value)

        # FHE_init
        # self._spl_so_face._FHE_init = self._spl_so_face.FHE_init
        self._spl_so_face.FHE_init.argtypes = [c_int]
        self._spl_so_face.FHE_init.restype = POINTER(c_uint8)
        self._spl_so_face.handle = self._spl_so_face.FHE_init(self._logging_level.value)

        self._spl_so_face.privid_initialize_session_join.argtypes = [POINTER(c_void_p), c_void_p]
        self._spl_so_face.privid_initialize_session_join.restype = c_bool
        self._spl_so_face.new_handle = c_void_p()

        self._spl_so_face.privid_initialize_session_join(byref(self._spl_so_face.new_handle), self._spl_so_face.handle)

        # FHE_configure_url
        # self._spl_so_face.FHE_configure_url = self._spl_so_face.FHE_configure_url
        self._spl_so_face.FHE_configure_url.argtypes = [
            POINTER(c_uint8), c_int, c_char_p, c_int]
        self._spl_so_face.FHE_configure_url.restype = c_uint8

        # privid_set_configuration
        self._spl_so_face.privid_set_configuration.argtypes = [c_void_p, c_char_p, c_int]
        self._spl_so_face.privid_set_configuration.restype = c_bool

        # configure_url , API key and storage location
        self._spl_so_face.FHE_configure_url(self._spl_so_face.handle, c_int32(46),
                                            c_char_p(self._api_key),
                                            c_int32(len(self._api_key)))

        self._spl_so_face.FHE_configure_url(self._spl_so_face.handle, c_int32(42),
                                            c_char_p(self._server_url),
                                            c_int32(len(self._server_url)))

        # Configure parameters 
        if self._config_object and self._config_object.get_config_param():
            config_dict = json.loads(self._config_object.get_config_param())
            config_dict["cache_type"] = self._cache_type.value
            config_dict["local_storage_path"] = self._local_storage_path
            config_dict = json.dumps(config_dict)
            c_config_param = c_char_p(bytes(config_dict, 'utf-8'))
            c_config_param_len = c_int(len(config_dict))
            self._spl_so_face.privid_set_configuration(self._spl_so_face.new_handle, c_config_param, c_config_param_len)
        # self._spl_so_face.privid_set_configuration.restype = c_int
        # FHE_close
        # self._spl_so_face.FHE_close = self._spl_so_face.FHE_close
        self._spl_so_face.FHE_close.argtypes = [POINTER(c_uint8)]
        self._spl_so_face.FHE_close.restype = c_int

        # privid_enroll_onefa
        # self._spl_so_face.privid_enroll_onefa = self._spl_so_face.privid_enroll_onefa
        self._spl_so_face.privid_enroll_onefa.argtypes = [c_void_p, c_char_p, c_int, POINTER(
            c_uint8), c_int, c_int, c_int, c_int, POINTER(c_float), POINTER(c_int), c_bool, POINTER(c_uint8),
                                                          POINTER(c_int), POINTER(c_char_p),
                                                          POINTER(c_int)]
        self._spl_so_face.privid_enroll_onefa.restype = c_int

        # privid_face_predict_onefa
        # self._spl_so_face.privid_face_predict_onefa = self._spl_so_face.privid_face_predict_onefa
        self._spl_so_face.privid_face_predict_onefa.argtypes = [c_void_p, c_char_p, c_int, POINTER(
            c_uint8), c_int, c_int, c_int, c_int, POINTER(c_float), POINTER(c_int), c_bool, POINTER(c_uint8),
                                                                POINTER(c_int), POINTER(c_char_p),
                                                                POINTER(c_int)]
        self._spl_so_face.privid_face_predict_onefa.restype = c_int

        # is_valid
        # self._spl_so_face.is_valid = self._spl_so_face.is_valid
        self._spl_so_face.is_valid.argtypes = [POINTER(c_uint8), c_bool, POINTER(
            c_uint8), c_int, c_int, POINTER(c_uint8), POINTER(c_int), POINTER(c_char_p), POINTER(c_int),
                                               POINTER(c_char_p), c_int]
        self._spl_so_face.is_valid.restype = c_int

        # FHE_delete
        # self._spl_so_face.privid_user_delete = self._spl_so_face.privid_user_delete
        self._spl_so_face.privid_user_delete.argtypes = [c_void_p, POINTER(c_char), c_int, POINTER(
            c_char), c_int, POINTER(c_char_p), POINTER(c_int)]
        self._spl_so_face.privid_user_delete.restype = c_int

        # FHE_free_api_memory
        # self._spl_so_face.FHE_free_api_memory = self._spl_so_face.FHE_free_api_memory
        self._spl_so_face.FHE_free_api_memory.argtypes = [POINTER(c_char_p)]

        # FHE_compare_files
        # self._spl_so_face.privid_face_compare_files = self._spl_so_face.privid_face_compare_files
        self._spl_so_face.privid_face_compare_files.argtypes = [c_void_p, c_float, c_char_p, c_int,
                                                                POINTER(c_uint8), c_int, c_int, c_int,
                                                                POINTER(c_uint8), c_int, c_int, c_int,
                                                                POINTER(c_char_p),
                                                                POINTER(c_int)]
        self._spl_so_face.privid_face_compare_files.restype = c_int

        # privid_validate
        self._spl_so_face.privid_validate.argtypes = [
            c_void_p, POINTER(c_uint8), c_int, c_int,
            c_char_p, c_int, POINTER(c_char_p), POINTER(c_int)]
        self._spl_so_face.privid_validate.restype = c_bool

        # privid_estimate_age
        self._spl_so_face.privid_estimate_age.argtypes = [
            c_void_p, POINTER(c_uint8), c_int, c_int,
            c_char_p, c_int, POINTER(c_char_p), POINTER(c_int)]
        self._spl_so_face.privid_estimate_age.restype = c_bool

        # privid_face_iso
        self._spl_so_face.privid_face_iso.argtypes = [
            c_void_p, POINTER(c_uint8), c_int, c_int,
            c_char_p, c_int, POINTER(c_char_p), POINTER(c_int),
            POINTER(POINTER(c_ubyte)), POINTER(c_int)]
        self._spl_so_face.privid_estimate_age.restype = c_bool

        # privid_set_billing_record_threshold
        self._spl_so_face.privid_set_billing_record_threshold.argtypes = [c_void_p, c_char_p, c_int]
        self._spl_so_face.privid_set_billing_record_threshold.restype = c_bool

        if self._config_object and self._config_object.get_config_billing_param():
            c_config_param = c_char_p(bytes(self._config_object.get_config_billing_param(), 'utf-8'))
            c_config_param_len = c_int(len(self._config_object.get_config_billing_param()))
            self._spl_so_face.privid_set_billing_record_threshold(self._spl_so_face.new_handle, c_config_param,
                                                                  c_config_param_len)

    def is_valid(self, image_data: np.array, is_enroll: bool = False, config_object: ConfigObject = None) -> Any:
        try:
            img = image_data
            im_width = img.shape[1]
            im_height = img.shape[0]
            p_buffer_images_in = img.flatten()
            c_p_buffer_images_in = p_buffer_images_in.ctypes.data_as(
                POINTER(c_uint8))
            p_buffer_embeddings_out = np.zeros(224 * 224 * 4, dtype=np.int8)
            c_p_buffer_embeddings_out = p_buffer_embeddings_out.ctypes.data_as(
                POINTER(c_uint8))

            result_out = np.zeros(1, dtype=np.int32)
            c_result_out = result_out.ctypes.data_as(POINTER(ctypes.c_int32))

            buffer_embeddings_out_length = np.zeros(1, dtype=np.int32)
            c_buffer_embeddings_out_length = buffer_embeddings_out_length.ctypes.data_as(POINTER(ctypes.c_int32))

            c_result = c_char_p()
            self._spl_so_face.is_valid(self._spl_so_face.handle, c_bool(is_enroll), c_p_buffer_images_in,
                                       c_int(im_width),
                                       c_int(im_height), c_p_buffer_embeddings_out,
                                       c_buffer_embeddings_out_length, byref(c_result), c_result_out,
                                       c_char_p(), c_int(0))

            len_ = np.fromiter(c_result_out[:1], dtype=np.uint32, count=-1)[0]
            output_json = c_result.value[:len_].decode()
            self._spl_so_face.FHE_free_api_memory(byref(c_result))
            output = json.loads(output_json)
            return output
        except Exception as e:
            print(e)
            return False

    def is_valid_without_age(self, image_data: np.array, config_object: ConfigObject = None) -> Any:
        try:
            img = image_data
            im_width = img.shape[1]
            im_height = img.shape[0]

            p_buffer_images_in = img.flatten()
            c_p_buffer_images_in = p_buffer_images_in.ctypes.data_as(POINTER(c_uint8))

            c_result = c_char_p()
            c_result_len = c_int()
            if config_object and config_object.get_config_param():
                c_config_param = c_char_p(bytes(config_object.get_config_param(), 'utf-8'))
                c_config_param_len = c_int(len(config_object.get_config_param()))
            else:
                c_config_param = c_char_p(bytes("", 'utf-8'))
                c_config_param_len = c_int(0)
            self._spl_so_face.privid_validate(
                self._spl_so_face.new_handle, c_p_buffer_images_in, c_int(im_width), c_int(im_height),
                c_config_param, c_config_param_len,
                byref(c_result), byref(c_result_len))

            if not c_result.value or not c_result_len.value:
                raise Exception("Something went wrong. Couldn't process the image for is_valid API. ")
            output_json = c_result.value[:c_result_len.value].decode()
            self._spl_so_face.FHE_free_api_memory(c_result)

            output = json.loads(output_json)
            return output
        except Exception as e:
            print(e)
            return False

    def estimate_age(self, image_data: np.array, config_object: ConfigObject = None) -> Any:
        try:
            img = image_data
            im_width = img.shape[1]
            im_height = img.shape[0]

            p_buffer_images_in = img.flatten()
            c_p_buffer_images_in = p_buffer_images_in.ctypes.data_as(POINTER(c_uint8))

            c_result = c_char_p()
            c_result_len = c_int()
            if config_object and config_object.get_config_param():
                c_config_param = c_char_p(bytes(config_object.get_config_param(), 'utf-8'))
                c_config_param_len = c_int(len(config_object.get_config_param()))
            else:
                c_config_param = c_char_p(bytes("", 'utf-8'))
                c_config_param_len = c_int(0)
            self._spl_so_face.privid_estimate_age(
                self._spl_so_face.new_handle, c_p_buffer_images_in, c_int(im_width), c_int(im_height),
                c_config_param, c_config_param_len,
                byref(c_result), byref(c_result_len))

            if not c_result.value or not c_result_len.value:
                raise Exception("Something went wrong. Couldn't process the image for estimate_age API. ")
            output_json = c_result.value[:c_result_len.value].decode()
            self._spl_so_face.FHE_free_api_memory(c_result)

            output = json.loads(output_json)
            return output
        except Exception as e:
            print(e)
            return False

    def get_iso_face(self, image_data: np.array, config_object: ConfigObject = None) -> Any:
        try:
            img = image_data
            im_width = img.shape[1]
            im_height = img.shape[0]

            p_buffer_images_in = img.flatten()
            c_p_buffer_images_in = p_buffer_images_in.ctypes.data_as(POINTER(c_uint8))

            c_result = c_char_p()
            c_result_len = c_int()
            c_iso_image_len = c_int()

            c_iso_image = POINTER(c_ubyte)()

            if config_object and config_object.get_config_param():
                c_config_param = c_char_p(bytes(config_object.get_config_param(), 'utf-8'))
                c_config_param_len = c_int(len(config_object.get_config_param()))
            else:
                c_config_param = c_char_p(bytes("", 'utf-8'))
                c_config_param_len = c_int(0)

            self._spl_so_face.privid_face_iso(
                self._spl_so_face.new_handle, c_p_buffer_images_in, c_int(im_width), c_int(im_height),
                c_config_param, c_config_param_len,
                byref(c_result), byref(c_result_len), byref(c_iso_image), byref(c_iso_image_len))

            if not c_result.value or not c_result_len.value:
                raise Exception("Something went wrong. Couldn't process the image for get_iso_face API. ")
            output_json = c_result.value[:c_result_len.value].decode()
            output_json = json.loads(output_json)
            self._spl_so_face.FHE_free_api_memory(c_result)
            if c_iso_image_len.value and "iso_image_width" in output_json and "iso_image_height" in output_json:
                output_json["image"] = Image.fromarray(np.uint8(np.reshape(c_iso_image[:c_iso_image_len.value], (
                    output_json.get("iso_image_height", 0), output_json.get("iso_image_width", 0),
                    output_json.get("iso_image_channels", 0))))).convert(
                    "RGBA" if output_json.get("iso_image_channels", 0) == 4 else "RGB")
            else:
                # Empty Image
                output_json["image"] = Image.new("RGB", (800, 1280), (255, 255, 255))
            return output_json
        except Exception as e:
            print(e)
            return False

    def delete(self, uuid: str) -> Any:
        uuid = bytes(uuid, 'utf-8')
        p_buffer_result = c_char_p()
        p_buffer_result_length = c_int()
        c_config_param = c_char_p(bytes("", 'utf-8'))
        c_config_param_len = c_int(0)
        self._spl_so_face.privid_user_delete(self._spl_so_face.new_handle, c_config_param, c_config_param_len,
                                             c_char_p(uuid),
                                             c_int(
                                                 len(uuid)), byref(p_buffer_result),
                                             byref(p_buffer_result_length))

        len_ = p_buffer_result_length.value
        output_json_str = p_buffer_result.value[:len_].decode()

        self._spl_so_face.FHE_free_api_memory(byref(p_buffer_result))

        if output_json_str is not None and len(output_json_str) > 0:
            output = json.loads(output_json_str)
            return output
        else:
            return False

    def compare_files(self, left_image: np.array, right_image: np.array, config_object: ConfigObject = None) -> Any:
        fudge_factor = 0.0
        try:
            left_img_data_buffer = left_image.flatten()
            right_img_data_buffer = right_image.flatten()

            left_c_img_data_buffer = left_img_data_buffer.ctypes.data_as(
                POINTER(c_uint8))
            right_c_img_data_buffer = right_img_data_buffer.ctypes.data_as(
                POINTER(c_uint8))

            lim_height, lim_width, _ = left_image.shape
            lim_size = left_image.shape[1] * left_image.shape[0] * left_image.shape[2]

            rim_height, rim_width, _ = right_image.shape
            rim_size = right_image.shape[1] * right_image.shape[0] * right_image.shape[2]

            p_buffer_result = c_char_p()
            p_buffer_result_length = c_int()
            if config_object and config_object.get_config_param():
                c_config_param = c_char_p(bytes(config_object.get_config_param(), 'utf-8'))
                c_config_param_len = c_int(len(config_object.get_config_param()))
            else:
                c_config_param = c_char_p(bytes("", 'utf-8'))
                c_config_param_len = c_int(0)
            self._spl_so_face.privid_face_compare_files(self._spl_so_face.new_handle, c_float(fudge_factor),
                                                        c_config_param,
                                                        c_config_param_len,
                                                        left_c_img_data_buffer,
                                                        c_int(lim_size), c_int(lim_width), c_int(lim_height),
                                                        right_c_img_data_buffer, c_int(rim_size),
                                                        c_int(rim_width), c_int(rim_height),
                                                        byref(p_buffer_result), byref(p_buffer_result_length))
            if not p_buffer_result.value or not p_buffer_result_length.value:
                raise Exception("Something went wrong. Please validate the images using isvalid function")
            len_ = p_buffer_result_length.value
            output_json_str = p_buffer_result.value[:len_].decode()

            self._spl_so_face.FHE_free_api_memory(byref(p_buffer_result))
            if output_json_str is not None and len(output_json_str) > 0:
                output = json.loads(output_json_str)
                output["status"] = output["result"]
                return output
            else:
                return False
        except Exception as e:
            print(e)
            return False

    def enroll(self, image_data: np.array, config_object: ConfigObject = None) -> Any:
        im_count = 1
        try:
            img_data = image_data
            im_height, im_width, im_channel = img_data.shape

            p_buffer_images_in = img_data
            c_p_buffer_images_in = p_buffer_images_in.ctypes.data_as(
                POINTER(c_uint8))
            im_size = im_height * im_width * im_channel
            p_buffer_embeddings_out = np.zeros(
                4 * self._embedding_length * self._num_embeddings, dtype=np.float32)
            c_p_buffer_embeddings_out = p_buffer_embeddings_out.ctypes.data_as(
                POINTER(c_float))

            augmented_images = np.zeros(self._aug_size, dtype=np.int8)
            c_augmented_images = augmented_images.ctypes.data_as(
                POINTER(c_uint8))

            result_out = np.zeros(1, dtype=np.int32)
            c_result_out = result_out.ctypes.data_as(POINTER(ctypes.c_int32))
            c_result = c_char_p()
            emb_out_length = np.zeros(1, dtype=np.int32)
            c_emb_out_length = emb_out_length.ctypes.data_as(POINTER(ctypes.c_int32))

            augmented_images_length = np.zeros(1, dtype=np.int32)
            c_augmented_images_length = augmented_images_length.ctypes.data_as(POINTER(ctypes.c_int32))
            if config_object and config_object.get_config_param():
                c_config_param = c_char_p(bytes(config_object.get_config_param(), 'utf-8'))
                c_config_param_len = c_int(len(config_object.get_config_param()))
            else:
                c_config_param = c_char_p(bytes("", 'utf-8'))
                c_config_param_len = c_int(0)
            self._spl_so_face.privid_enroll_onefa(self._spl_so_face.new_handle, c_config_param, c_config_param_len,
                                                  c_p_buffer_images_in,
                                                  c_int(im_count), c_int(im_size), c_int(im_width),
                                                  c_int(im_height),
                                                  c_p_buffer_embeddings_out, c_emb_out_length, c_bool(True),
                                                  c_augmented_images, c_augmented_images_length,
                                                  byref(c_result), c_result_out)

            len_ = np.fromiter(c_result_out[:1], dtype=np.uint32, count=-1)[0]
            output_json_str = c_result.value[:len_].decode()
            self._spl_so_face.FHE_free_api_memory(byref(c_result))
            if output_json_str is not None and len(output_json_str) > 0:
                output = json.loads(output_json_str)
                return output
            return False
        except Exception as e:
            print("Error :", e)
            return False

    def predict(self, image_data: np.array, config_object: ConfigObject = None) -> Any:
        im_count = 1
        try:
            img_data = image_data
            im_height, im_width, im_channel = img_data.shape

            p_buffer_images_in = img_data
            c_p_buffer_images_in = p_buffer_images_in.ctypes.data_as(
                POINTER(c_uint8))
            im_size = im_height * im_width * im_channel
            p_buffer_embeddings_out = np.zeros(
                4 * self._embedding_length * self._num_embeddings, dtype=np.float32)
            c_p_buffer_embeddings_out = p_buffer_embeddings_out.ctypes.data_as(
                POINTER(c_float))

            augmented_images = np.zeros(self._aug_size, dtype=np.int8)
            c_augmented_images = augmented_images.ctypes.data_as(
                POINTER(c_uint8))

            result_out = np.zeros(1, dtype=np.int32)
            c_result_out = result_out.ctypes.data_as(POINTER(ctypes.c_int32))

            emb_out_lenght = np.zeros(1, dtype=np.int32)
            emb_out_lenght = emb_out_lenght.ctypes.data_as(POINTER(ctypes.c_int32))
            c_result = c_char_p()
            augmented_images_length = np.zeros(1, dtype=np.int32)
            c_augmented_images_length = augmented_images_length.ctypes.data_as(POINTER(ctypes.c_int32))
            if config_object and config_object.get_config_param():
                c_config_param = c_char_p(bytes(config_object.get_config_param(), 'utf-8'))
                c_config_param_len = c_int(len(config_object.get_config_param()))
            else:
                c_config_param = c_char_p(bytes("", 'utf-8'))
                c_config_param_len = c_int(0)
            self._spl_so_face.privid_face_predict_onefa(self._spl_so_face.new_handle, c_config_param,
                                                        c_config_param_len,
                                                        c_p_buffer_images_in,
                                                        c_int(im_count), c_int(im_size), c_int(im_width),
                                                        c_int(im_height),
                                                        c_p_buffer_embeddings_out, emb_out_lenght, c_bool(True),
                                                        c_augmented_images, c_augmented_images_length, byref(c_result),
                                                        c_result_out)
            len_ = np.fromiter(c_result_out[:1], dtype=np.uint32, count=-1)[0]
            output_json_str = c_result.value[:len_].decode()
            self._spl_so_face.FHE_free_api_memory(byref(c_result))
            if output_json_str is not None and len(output_json_str) > 0:
                output = json.loads(output_json_str)
                return output
            return False
        except Exception as e:
            print("Error :", e)
            return False



from enum import Enum, EnumMeta

import numpy as np
from PIL import Image


def image_path_to_array(image_path: str, input_format: str) -> np.ndarray:
    image = Image.open(image_path).convert(input_format.upper())
    return np.array(image)


class Point(object):
    """Creates a point on a coordinate plane with values x and y.
    """

    def __init__(self, x=None, y=None):
        """
        Point object having the x and y coordinates
        """
        self._x = x
        self._y = y

    @property
    def x(self):
        """
        Returns the x coordinate of the object
        """
        return self._x

    @property
    def y(self):
        """
        Returns the y coordinate of the object
        """
        return self._y

    def __str__(self):
        return "Point(%s,%s)" % (self._x, self._y)


class BoundingBox(object):

    def __init__(self, top_left_coordinate=None, bottom_right_coordinate=None):
        """
        Bounding box class for capturing the coordinates of the detected image
        """
        self._bottom_right_coordinate = Point(None, None)
        self._top_left_coordinate = Point(None, None)
        if bottom_right_coordinate:
            self._bottom_right_coordinate = Point(bottom_right_coordinate.get("x", None), bottom_right_coordinate.get("y", None))
        if top_left_coordinate:
            self._top_left_coordinate = Point(top_left_coordinate.get("x", None), top_left_coordinate.get("y", None))

    @property
    def bottom_right_coordinate(self) -> Point:
        """
        Returns the bottom right coordinate

        Returns
        -------
        Point
            x: float

            y: float

        """
        return self._bottom_right_coordinate

    @property
    def top_left_coordinate(self) -> Point:
        """
        Returns the top left coordinate

        Returns
        -------
        Point
            x: float

            y: float

        """
        return self._top_left_coordinate

    @bottom_right_coordinate.setter
    def bottom_right_coordinate(self, value):
        self._bottom_right_coordinate = value

    @top_left_coordinate.setter
    def top_left_coordinate(self, value):
        self._top_left_coordinate = value


class FaceValidationCodeMeta(EnumMeta):
    def __contains__(cls, item):
        return item in [v.value for v in cls.__members__.values()]


class FaceValidationCode(Enum, metaclass=FaceValidationCodeMeta):
    InvalidImage = -100
    NoFace = -1
    ValidBiometric = 0
    ImageSpoof = 1
    VideoSpoof = 2
    TooClose = 3
    TooFaraway = 4
    TooFarToRight = 5
    TooFarToLeft = 6
    TooFarUp = 7
    TooFarDown = 8
    TooBlurry = 9
    GlassesOn = 10
    MaskOn = 11
    ChinTooFarLeft = 12
    ChinTooFarRight = 13
    ChinTooFarUp = 14
    ChinTooFarDown = 15



from enum import Enum


class CacheType(str, Enum):
    """
    Cache Type defines whether you want to store the data offline

    off - No cache

    on - Cache set in the local storage path

    """
    ON = "basic"
    OFF = "nocache"



class Message:
    def __init__(self):
        self.IS_VALID_ERROR = "Unknown error while validating the image"
        self.AGE_ESTIMATE_ERROR = "Unknown error while predicting the age."
        self.EXCEPTION_ERROR_ENROLL = "Something went wrong while doing enroll."
        self.EXCEPTION_ERROR_GET_ISO_FACE = "ISO face validation failed."
        self.EXCEPTION_ERROR_PREDICT = "Something went wrong while doing predict."
        self.EXCEPTION_ERROR_DELETE = "Something went wrong while doing delete."
        self.EXCEPTION_ERROR_COMPARE = "Something went wrong while doing compare."
        self.ENROLL_PROCESSED = "Enroll request successfully processed, Waiting for server response."
        self.PREDICT_PROCESSED = "Predict request successfully processed, Waiting for server response."

        self.APP_MESSAGES = {
            0: "Valid Image",
            1: "Error Description: Face is an image of an image (spoof). Please only provide live facial image(s).",
            2: "Error Description: Face is an image of a video (spoof). Please only provide live facial image(s).",
            3: "Error Description: Face in image is too close to the camera. Please move away from the camera.",
            4: "Error Description: Face in image is too far away.",
            5: "Error Description: Face in image is too far to the right.",
            6: "Error Description: Face in image is too far to the left.",
            7: "Error Description: Face in image is too high.",
            8: "Error Description: Face in image is too low.",
            9: "Error Description: Face in image is too blurry.",
            10: "Error Description: Please remove eyeglasses during registration.",
            11: "Error Description:  Please remove face mask  during registration. ",
            12: "Head in image turned too far toward the left/right. Please face the camera",
            13: "Head in image turned too far toward the up/down. Please face the camera",
            14: "UNUSED ERROR CODE",
            15: "UNUSED ERROR CODE",
            16: "Error Description: No face found in image.",
            17: "Error Description: API Error",
            18:	"Error Description: Local Storage Error",
            19: "Error Description: Memory Error",
            100: "Successfully registered",
            101: "Error Description: Image file does not exist.",
            102: "Error Description: Input image quality is low.",
            103: "Error Description: There is an error in endpoint.",
            104: "Factor object successfully created.",
            105: "Error Description: Something went wrong while initializing.",
            106: self.IS_VALID_ERROR,
            107: self.EXCEPTION_ERROR_ENROLL,
            108: self.EXCEPTION_ERROR_PREDICT,
            109: "Error Description: Incorrect Usage."
        }

    def get_message(self, code):
        return self.APP_MESSAGES[code]



import json
import sys
from enum import Enum, EnumMeta
from typing import Dict


class __PARAMETERSMETA(EnumMeta):
    def __contains__(cls, item):
        return item in [v for v in cls.__members__.values()]


class PARAMETERS(str, Enum, metaclass=__PARAMETERSMETA):
    """PARAMETERS contains the valid key values for configuring additional parameters in the factor processor.
    """
    INPUT_IMAGE_FORMAT = "input_image_format"
    CONTEXT_STRING = "context_string"
    CONF_FAST_PROCESS = "conf_fast_process"
    INPUT_TYPE = "input_type"
    FACE_THRESHOLDS_REM_BAD_EMB = "face_thresholds_rem_bad_emb"
    BLUR_THRESHOLD_DOC_LEVEL_1 = "blur_threshold_doc_level_1"
    BLUR_THRESHOLD_DOC_LEVEL_2 = "blur_threshold_doc_level_2"
    BLUR_THRESHOLD_ENROLL_PRED = "blur_threshold_enroll_pred"
    THRESHOLD_PROFILE_ENROLL = "threshold_profile_enroll"
    THRESHOLD_PROFILE_PREDICT = "threshold_profile_predict"
    THRESHOLD_VERTICAL_ENROLL = "threshold_vertical_enroll"
    THRESHOLD_VERTICAL_PREDICT = "threshold_vertical_predict"
    THRESHOLD_USER_RIGHT = "threshold_user_right"
    THRESHOLD_USER_LEFT = "threshold_user_left"
    THRESHOLD_USER_TOO_FAR = "threshold_user_too_far"
    THRESHOLD_USER_TOO_CLOSE = "threshold_user_too_close"
    IMAGE_BORDER = "image_border"
    IMAGE_PRE_PROC = "image_pre_proc"
    THRESHOLD_GLASS = "threshold_glass"
    THRESHOLD_MASK = "threshold_mask"
    FACE_THRESHOLD_RIGHT = "face_threshold_right"
    FACE_THRESHOLD_LEFT = "face_threshold_left"
    FACE_THRESHOLD_VERTICAL = "face_threshold_vertical"
    CONF_SCORE_THR_ENROLL = "conf_score_thr_enroll"
    CONF_SCORE_THR_PREDICT = "conf_score_thr_predict"
    MIN_DOCUMENT_BORDER = "min_document_border"
    DISALLOWED_RESULTS = "disallowed_results"
    ALLOWED_RESULTS = "allowed_results"
    DOCUMENT_FACE_CHECK_VALIDITY = "document_face_check_validity"
    DOCUMENT_CHECK_VALIDITY = "document_check_validity"
    DOCUMENT_FACE_PREDICT = "document_face_predict"
    ENABLE_DOC_PERSPECTIVE_CORRECTION = "enable_doc_perspective_correction"
    ENROLL_ALLOW_EYE_GLASS = "enroll_allow_eye_glass"
    ORIENTATION_ID_VALUE = "orientation_id_value"
    FACE_DETECT_PREFERRED_SIZE = "face_detect_preferred_size"
    FACE_DETECT_MAX_OUT_IMAGE_SIZE = "face_detect_max_out_image_size"
    SEND_ORIGINAL_IMAGES = "send_original_images"

    # BILLING PARAMETERS
    ENROLL_RESERVATION_CALLS = "enroll"
    ISVALID_RESERVATION_CALLS = "is_valid"
    PREDICT_RESERVATION_CALLS = "predict"
    DOC_FRONT_RESERVATION_CALLS = "document_model"
    DOC_BACK_RESERVATION_CALLS = "document_model"
    DELETE_RESERVATION_CALLS = "delete"
    COMPARE_RESERVATION_CALLS = "compare_files"
    FACE_RESERVATION_CALLS = "faces"
    ESTIMATE_AGE_RESERVATION_CALLS = "estimate_age"
    FACE_ISO_RESERVATION_CALLS = "face_iso"


class ParameterValidator:
    def __init__(self):
        self.__parameter = {}
        self.__populate_parameters()
        self.__billing_reservation_parameters = [PARAMETERS.ENROLL_RESERVATION_CALLS,
                                                 PARAMETERS.ISVALID_RESERVATION_CALLS,
                                                 PARAMETERS.PREDICT_RESERVATION_CALLS,
                                                 PARAMETERS.DOC_FRONT_RESERVATION_CALLS,
                                                 PARAMETERS.DOC_BACK_RESERVATION_CALLS,
                                                 PARAMETERS.DELETE_RESERVATION_CALLS,
                                                 PARAMETERS.COMPARE_RESERVATION_CALLS,
                                                 PARAMETERS.FACE_RESERVATION_CALLS,
                                                 PARAMETERS.ESTIMATE_AGE_RESERVATION_CALLS,
                                                 PARAMETERS.FACE_ISO_RESERVATION_CALLS]

    def __populate_parameters(self):
        self.__parameter[PARAMETERS.INPUT_IMAGE_FORMAT] = self.Parameter(
            name=PARAMETERS.INPUT_IMAGE_FORMAT, _type="SET", valid_set=["rgb", "rgba", "bgr"])
        self.__parameter[PARAMETERS.CONTEXT_STRING] = self.Parameter(name=PARAMETERS.CONTEXT_STRING,
                                                                     _type="SET",
                                                                     valid_set=["enroll", "predict"])
        self.__parameter[PARAMETERS.CONF_FAST_PROCESS] = self.Parameter(name=PARAMETERS.CONF_FAST_PROCESS,
                                                                        _type="BOOL")
        self.__parameter[PARAMETERS.INPUT_TYPE] = self.Parameter(name=PARAMETERS.INPUT_TYPE, _type="SET",
                                                                 valid_set=["face", "document-id",
                                                                            "document-barcode"])
        self.__parameter[PARAMETERS.FACE_THRESHOLDS_REM_BAD_EMB] = self.Parameter(
            name=PARAMETERS.FACE_THRESHOLDS_REM_BAD_EMB, _type="NUMBER", min_value=0, max_value=2)
        self.__parameter[PARAMETERS.BLUR_THRESHOLD_DOC_LEVEL_1] = self.Parameter(
            name=PARAMETERS.BLUR_THRESHOLD_DOC_LEVEL_1, _type="NUMBER", min_value=0, max_value=10000)
        self.__parameter[PARAMETERS.BLUR_THRESHOLD_DOC_LEVEL_2] = self.Parameter(
            name=PARAMETERS.BLUR_THRESHOLD_DOC_LEVEL_2, _type="NUMBER", min_value=0, max_value=10000)
        self.__parameter[PARAMETERS.BLUR_THRESHOLD_ENROLL_PRED] = self.Parameter(
            name=PARAMETERS.BLUR_THRESHOLD_ENROLL_PRED, _type="NUMBER", min_value=0, max_value=10000)
        self.__parameter[PARAMETERS.THRESHOLD_PROFILE_ENROLL] = self.Parameter(
            name=PARAMETERS.THRESHOLD_PROFILE_ENROLL, _type="NUMBER", min_value=-0.1, max_value=2)
        self.__parameter[PARAMETERS.THRESHOLD_PROFILE_PREDICT] = self.Parameter(
            name=PARAMETERS.THRESHOLD_PROFILE_PREDICT, _type="NUMBER", min_value=-0.1, max_value=2)
        self.__parameter[PARAMETERS.THRESHOLD_VERTICAL_ENROLL] = self.Parameter(
            name=PARAMETERS.THRESHOLD_VERTICAL_ENROLL, _type="NUMBER", min_value=-0.1, max_value=2)
        self.__parameter[PARAMETERS.THRESHOLD_VERTICAL_PREDICT] = self.Parameter(
            name=PARAMETERS.THRESHOLD_VERTICAL_PREDICT, _type="NUMBER", min_value=-0.1, max_value=2)
        self.__parameter[PARAMETERS.THRESHOLD_USER_RIGHT] = self.Parameter(
            name=PARAMETERS.THRESHOLD_USER_RIGHT, _type="NUMBER", min_value=-0.1, max_value=2)
        self.__parameter[PARAMETERS.THRESHOLD_USER_LEFT] = self.Parameter(
            name=PARAMETERS.THRESHOLD_USER_LEFT, _type="NUMBER", min_value=-0.1, max_value=2)
        self.__parameter[PARAMETERS.THRESHOLD_USER_TOO_FAR] = self.Parameter(
            name=PARAMETERS.THRESHOLD_USER_TOO_FAR, _type="NUMBER", min_value=-0.1, max_value=2)
        self.__parameter[PARAMETERS.THRESHOLD_USER_TOO_CLOSE] = self.Parameter(
            name=PARAMETERS.THRESHOLD_USER_TOO_CLOSE, _type="NUMBER", min_value=-0.1, max_value=2)
        self.__parameter[PARAMETERS.IMAGE_BORDER] = self.Parameter(name=PARAMETERS.IMAGE_BORDER,
                                                                   _type="NUMBER", min_value=0, max_value=0.1)
        self.__parameter[PARAMETERS.IMAGE_PRE_PROC] = self.Parameter(name=PARAMETERS.IMAGE_PRE_PROC,
                                                                     _type="SET",
                                                                     valid_set=["zoom_pan", "rotate90",
                                                                                "rotate180", "rotate270", "blur",
                                                                                "fliplr", "none"])
        self.__parameter[PARAMETERS.THRESHOLD_GLASS] = self.Parameter(name=PARAMETERS.THRESHOLD_GLASS,
                                                                      _type="NUMBER", min_value=-0.1,
                                                                      max_value=2)
        self.__parameter[PARAMETERS.THRESHOLD_MASK] = self.Parameter(name=PARAMETERS.THRESHOLD_MASK,
                                                                     _type="NUMBER", min_value=-0.1, max_value=2)
        self.__parameter[PARAMETERS.FACE_THRESHOLD_RIGHT] = self.Parameter(
            name=PARAMETERS.FACE_THRESHOLD_RIGHT, _type="ANY")
        self.__parameter[PARAMETERS.FACE_THRESHOLD_LEFT] = self.Parameter(
            name=PARAMETERS.FACE_THRESHOLD_LEFT, _type="ANY")
        self.__parameter[PARAMETERS.FACE_THRESHOLD_VERTICAL] = self.Parameter(
            name=PARAMETERS.FACE_THRESHOLD_VERTICAL, _type="ANY")
        self.__parameter[PARAMETERS.CONF_SCORE_THR_ENROLL] = self.Parameter(
            name=PARAMETERS.CONF_SCORE_THR_ENROLL, _type="NUMBER", min_value=-0.1, max_value=2)
        self.__parameter[PARAMETERS.CONF_SCORE_THR_PREDICT] = self.Parameter(
            name=PARAMETERS.CONF_SCORE_THR_PREDICT, _type="NUMBER", min_value=-0.1, max_value=2)
        self.__parameter[PARAMETERS.MIN_DOCUMENT_BORDER] = self.Parameter(
            name=PARAMETERS.MIN_DOCUMENT_BORDER, _type="ANY")
        self.__parameter[PARAMETERS.DISALLOWED_RESULTS] = self.Parameter(
            name=PARAMETERS.DISALLOWED_RESULTS, _type="ANY")
        self.__parameter[PARAMETERS.ALLOWED_RESULTS] = self.Parameter(name=PARAMETERS.DISALLOWED_RESULTS,
                                                                      _type="ANY")
        self.__parameter[PARAMETERS.DOCUMENT_FACE_CHECK_VALIDITY] = self.Parameter(
            name=PARAMETERS.DOCUMENT_FACE_CHECK_VALIDITY, _type="BOOL")
        self.__parameter[PARAMETERS.DOCUMENT_CHECK_VALIDITY] = self.Parameter(
            name=PARAMETERS.DOCUMENT_CHECK_VALIDITY, _type="BOOL")
        self.__parameter[PARAMETERS.DOCUMENT_FACE_PREDICT] = self.Parameter(
            name=PARAMETERS.DOCUMENT_FACE_PREDICT, _type="BOOL")
        self.__parameter[PARAMETERS.ENABLE_DOC_PERSPECTIVE_CORRECTION] = self.Parameter(
            name=PARAMETERS.ENABLE_DOC_PERSPECTIVE_CORRECTION, _type="BOOL")
        self.__parameter[PARAMETERS.ENROLL_ALLOW_EYE_GLASS] = self.Parameter(
            name=PARAMETERS.ENROLL_ALLOW_EYE_GLASS, _type="BOOL")
        self.__parameter[PARAMETERS.ORIENTATION_ID_VALUE] = self.Parameter(
            name=PARAMETERS.ORIENTATION_ID_VALUE, _type="ANY")
        self.__parameter[PARAMETERS.FACE_DETECT_PREFERRED_SIZE] = self.Parameter(
            name=PARAMETERS.FACE_DETECT_PREFERRED_SIZE, _type="ANY")
        self.__parameter[PARAMETERS.FACE_DETECT_MAX_OUT_IMAGE_SIZE] = self.Parameter(
            name=PARAMETERS.FACE_DETECT_MAX_OUT_IMAGE_SIZE, _type="ANY")
        self.__parameter[PARAMETERS.SEND_ORIGINAL_IMAGES] = self.Parameter(
            name=PARAMETERS.SEND_ORIGINAL_IMAGES, _type="BOOL")

        # BILLING PARAMETERS
        self.__parameter[PARAMETERS.ENROLL_RESERVATION_CALLS] = self.Parameter(
            name=PARAMETERS.ENROLL_RESERVATION_CALLS, _type="NUMBER", min_value=0, max_value=100000000)

        self.__parameter[PARAMETERS.ISVALID_RESERVATION_CALLS] = self.Parameter(
            name=PARAMETERS.ISVALID_RESERVATION_CALLS, _type="NUMBER", min_value=0, max_value=100000000)
        self.__parameter[PARAMETERS.PREDICT_RESERVATION_CALLS] = self.Parameter(
            name=PARAMETERS.PREDICT_RESERVATION_CALLS, _type="NUMBER", min_value=0, max_value=100000000)

        self.__parameter[PARAMETERS.DOC_FRONT_RESERVATION_CALLS] = self.Parameter(
            name=PARAMETERS.DOC_FRONT_RESERVATION_CALLS, _type="NUMBER", min_value=0, max_value=100000000)

        self.__parameter[PARAMETERS.DOC_BACK_RESERVATION_CALLS] = self.Parameter(
            name=PARAMETERS.DOC_BACK_RESERVATION_CALLS, _type="NUMBER", min_value=0, max_value=100000000)

        self.__parameter[PARAMETERS.DELETE_RESERVATION_CALLS] = self.Parameter(
            name=PARAMETERS.DELETE_RESERVATION_CALLS, _type="NUMBER", min_value=0, max_value=100000000)

        self.__parameter[PARAMETERS.COMPARE_RESERVATION_CALLS] = self.Parameter(
            name=PARAMETERS.COMPARE_RESERVATION_CALLS, _type="NUMBER", min_value=0, max_value=100000000)

        self.__parameter[PARAMETERS.FACE_RESERVATION_CALLS] = self.Parameter(
            name=PARAMETERS.FACE_RESERVATION_CALLS, _type="NUMBER", min_value=0, max_value=100000000)

        self.__parameter[PARAMETERS.ESTIMATE_AGE_RESERVATION_CALLS] = self.Parameter(
            name=PARAMETERS.ESTIMATE_AGE_RESERVATION_CALLS, _type="NUMBER", min_value=0, max_value=100000000)

        self.__parameter[PARAMETERS.FACE_ISO_RESERVATION_CALLS] = self.Parameter(
            name=PARAMETERS.FACE_ISO_RESERVATION_CALLS, _type="NUMBER", min_value=0, max_value=100000000)

    def validate(self, key, value):
        return self.__parameter[key].validate(value)

    def is_billing_parameter(self, key):
        return key in self.__billing_reservation_parameters

    class Parameter:
        def __init__(self, name=None, _type=None, min_value=None, max_value=None, valid_set=None):
            self.__name = name
            # ANY, BOOL, SET, NUMBER
            self.__type = _type
            self.__min_value = min_value
            self.__max_value = max_value
            self.__valid_set = valid_set

        def validate(self, value):
            if self.__type == "ANY":
                return True
            if self.__type == "BOOL":
                if value in [False, "True", "true", "false", "False", True]:
                    return True
                return False

            if self.__type == "SET":
                if value in self.__valid_set:
                    return True
                return False

            if self.__type == "NUMBER":
                if isinstance(value, (int, float)) and self.__min_value <= value <= self.__max_value:
                    return True
                return False

            return False


class ConfigObject:
    """Configuration Object class handles the parameters that are required to initialize the server with
    other fine controlling variables for factor processor.

    Parameters
    ----------
    config_param : Dict[PARAMETERS, str]
        Configuration parameters for changing the behaviour of face processing.
        Refer PARAMETERS class for valid values

    Returns
    -------
    ConfigObject
        Instance of the ConfigObject class.
    """

    def __init__(self, config_param: Dict[PARAMETERS, str] = None):
        if config_param is None:
            self._config_param = {}
        else:
            self._config_param = config_param
        self._validate_parameter = ParameterValidator()
        self.parse_config()

    def parse_config(self):
        try:
            for key, value in self._config_param.items():
                if key not in PARAMETERS:
                    raise ValueError("Invalid key '{}' in config parameters".format(key))

                if not self._validate_parameter.validate(key, value):
                    raise ValueError("Invalid key value pair\n'{}' : '{}'".format(key, value))
        except ValueError as exp:
            print("Config Error:", exp)
            sys.exit(1)

    def get_config_param(self):
        if len(self._config_param) == 0:
            return None
        config_param_dict = {}
        for key, value in self._config_param.items():
            if not self._validate_parameter.is_billing_parameter(key):
                config_param_dict[key.value] = value

        if len(config_param_dict) == 0:
            return None
        return json.dumps(config_param_dict)

    def get_config_billing_param(self):
        if len(self._config_param) == 0:
            return None
        config_billing_param = {}
        for key, value in self._config_param.items():
            if self._validate_parameter.is_billing_parameter(key):
                config_billing_param[key.value] = value

        if len(config_billing_param) == 0:
            return None
        return json.dumps(config_billing_param)

    def __str__(self):
        config_string = ""
        for key, value in self._config_param.items():
            config_string += "{}\t{}\n".format(key, value)
        return "\nPARAMETERS\n{}\n{}".format("-" * 10, config_string)

    @property
    def config_param(self) -> Dict[PARAMETERS, str]:
        """
        Returns the configuration parameter
        """
        return self._config_param

    @config_param.setter
    def config_param(self, value):
        self._config_param = value



"""Factor implements the functionalities of each factors to generate and verify Private IDs .
"""
import os
import pathlib
import platform
import sys
import traceback

import numpy as np

from .factor_modules.FaceModule import Face
from .helper.decorators import Singleton, deprecated
from .helper.messages import Message
from .helper.result_objects.compareResult import FaceCompareResult
from .helper.result_objects.deleteResult import FaceDeleteResult
from .helper.result_objects.enrollPredictResult import FaceEnrollPredictResult
from .helper.result_objects.faceValidationResult import FaceValidationResult
from .helper.result_objects.isValidDeprecatedResult import FaceIsValidDeprecatedResult
from .helper.result_objects.isoFaceResult import ISOFaceResult
from .helper.utils import image_path_to_array
from .settings.cacheType import CacheType
from .settings.configuration import ConfigObject, PARAMETERS
from .settings.loggingLevel import LoggingLevel
from .settings.supportedPlatforms import SupportedPlatforms


class FaceFactor(metaclass=Singleton):
    """The FaceFactor class implements the methods for enrolling and predicting the Face module as part of the
        Biometric Authentication.

        It exposes five methods as part of the interface:

        1. is_valid: Verifies the face of the user.
        2. estimate_age: Predicts the age of the face.
        3. compare: Compare two faces for verification.
        4. enroll: Enrolls the face of the user.
        5. predict: Predicts the face of the user.
        6. delete: Deletes the user from the system

        Parameters
        ----------
        api_key : str
            The API key for using the FaceFactor server.

        server_url : str
            The URL of the FaceFactor server.

        local_storage_path : str (optional)
            Absolute path to the local storage.

        logging_level : LoggingLevel (Optional)
            LoggingLevel needed while performing operation

        tf_num_thread: int (Optional)
            Number of thread to use for Tensorflow model inference

        cache_type: CacheType (Optional)
            To set the cache on / off

        config : ConfigObject (Optional)
            Configuration class object with parameters

        Returns
        -------
        FaceFactor
            Instance of the FaceFactor class.

        Methods
        -------
        is_valid
        estimate_age
        compare
        enroll
        predict
        delete
        get_iso_face
    """

    def __init__(self, api_key: str = None, server_url: str = None, local_storage_path: str = None,
                 logging_level: LoggingLevel = LoggingLevel.off, tf_num_thread: int = 0,
                 cache_type: CacheType = CacheType.OFF, config: ConfigObject = None):

        try:
            if platform.system() not in SupportedPlatforms.supportedOS.value:
                raise OSError("Invalid OS")
            if server_url is None and (os.environ.get('PI_SERVER_URL') is None or len(
                    os.environ.get('PI_SERVER_URL')) <= 0):
                raise ValueError("Server URL has to be configured")
            if api_key is None and (os.environ.get('PI_API_KEY') is None or len(
                    os.environ.get('PI_API_KEY')) <= 0):
                raise ValueError("API Key is required.")

            if tf_num_thread is None and (os.environ.get('PI_TF_NUM_THREAD') is None or len(
                    os.environ.get('PI_TF_NUM_THREAD')) <= 0):
                tf_num_thread = 0

            if tf_num_thread is None:
                self._tf_num_thread = int(os.environ.get('PI_TF_NUM_THREAD'))
            else:
                self._tf_num_thread = tf_num_thread
            if server_url is None:
                self._server_url = os.environ.get('PI_SERVER_URL')
            else:
                self._server_url = server_url
            if api_key is None:
                self._api_key = os.environ.get('PI_API_KEY')
            else:
                self._api_key = api_key
            if local_storage_path is None:
                self._local_storage_path = str(
                    pathlib.Path(__file__).parent.parent.joinpath("privateid_local_storage").resolve())
            else:
                self._local_storage_path = local_storage_path

            if self._server_url[-1] == "/":
                self._server_url = self._server_url[:-1]

            self._config_object = config
            self._logging_level = logging_level
            self._cache_type = cache_type
            self.face_factor = Face(api_key=self._api_key, server_url=self._server_url,
                                    local_storage_path=self._local_storage_path, logging_level=self._logging_level,
                                    tf_num_thread=self._tf_num_thread, cache_type=self._cache_type,
                                    config_object=self._config_object)
            self.message = Message()
        except ValueError as exp:
            print("Initialization Failed: {}\n".format(exp))
            print("Please refer to the usage documentation for setting up "
                  "Factor:: \nhttps://docs.private.id/cryptonets-python-sdk/usage.html")
            sys.exit(1)
        except OSError as exp:
            print("Initialization Failed: {}\n".format(exp))
            print("Please refer to the usage documentation for setting up "
                  "Factor:: \nhttps://docs.private.id/cryptonets-python-sdk/usage.html")
            sys.exit(1)

    def update_config(self, config):
        self.face_factor.update_config(config_object=config)

    @deprecated
    def is_valid_deprecated(self, image_path: str = None, image_data: np.array = None,
                            config: ConfigObject = None) -> FaceIsValidDeprecatedResult:
        """Check if the image is valid for using in the face recognition
        Deprecated: Use the is_valid instead as it improves the processing speed of is valid.

        Parameters
        ----------
        image_path
            Directory path to the image file

        config (Optional)
            Additional configuration parameters for the operation

        image_data (Optional)
            Image data in numpy format

        Returns
        -------
        FaceValidationResult
            status: int [0 if successful -1 if unsuccessful]

            message: str [Message from the operation]

            result: str [Result of the operation]

            age_factor: str [Predicted age of the image]

            output_image_data: any [Numpy RGB image data of cropped face]

        """
        try:
            if config is not None and PARAMETERS.INPUT_IMAGE_FORMAT in config.config_param:
                input_format = config.config_param[PARAMETERS.INPUT_IMAGE_FORMAT]
            elif self.config is not None and PARAMETERS.INPUT_IMAGE_FORMAT in self.config.config_param:
                input_format = self.config.config_param[PARAMETERS.INPUT_IMAGE_FORMAT]
            else:
                input_format = "rgb"
            if (image_path is not None and image_data is not None) or (image_path is None and image_data is None):
                return FaceIsValidDeprecatedResult(message="Specify either image_path or image_data")
            img_data = None
            if image_data is not None:
                if not isinstance(image_data, np.ndarray):
                    return FaceIsValidDeprecatedResult(message="Required numpy array in RGB/RGBA/BGR format")
                img_data = image_data
            if image_path is not None and len(image_path) > 0:
                if not os.path.exists(image_path):
                    return FaceIsValidDeprecatedResult(message=self.message.get_message(101))
                img_data = image_path_to_array(image_path, input_format=input_format)
            if img_data is None:
                return FaceIsValidDeprecatedResult(message=self.message.IS_VALID_ERROR)
            return self.face_factor.is_valid_deprecated(image_data=img_data, config_object=config)
        except Exception as e:
            print("Oops: {}\nTrace: {}".format(e, traceback.format_exc()))
            print("Issue Tracker:: \nhttps://github.com/prividentity/cryptonets-python-sdk/issues")
            return FaceIsValidDeprecatedResult(message=self.message.IS_VALID_ERROR)

    def is_valid(self, image_path: str = None, image_data: np.array = None,
                 config: ConfigObject = None) -> FaceValidationResult:
        """Check if the image is valid for using in the face recognition

        Parameters
        ----------
        image_path
            Directory path to the image file

        config (Optional)
            Additional configuration parameters for the operation

        image_data (Optional)
            Image data in numpy RGB format

        Returns
        -------
        FaceValidationResult
            error: int [0 if successful -1 if any error]

            message: str [Message from the operation]

            face_objects: List[FaceObjectResult]

        """
        try:
            if config is not None and PARAMETERS.INPUT_IMAGE_FORMAT in config.config_param:
                input_format = config.config_param[PARAMETERS.INPUT_IMAGE_FORMAT]
            elif self.config is not None and PARAMETERS.INPUT_IMAGE_FORMAT in self.config.config_param:
                input_format = self.config.config_param[PARAMETERS.INPUT_IMAGE_FORMAT]
            else:
                input_format = "rgb"
            if (image_path is not None and image_data is not None) or (image_path is None and image_data is None):
                return FaceValidationResult(message="Specify either image_path or image_data")
            img_data = None
            if image_data is not None:
                if not isinstance(image_data, np.ndarray):
                    return FaceValidationResult(message="Required numpy array in RGB/RGBA/BGR format")
                img_data = image_data
            if image_path is not None and len(image_path) > 0:
                if not os.path.exists(image_path):
                    return FaceValidationResult(message=self.message.get_message(101))
                img_data = image_path_to_array(image_path, input_format=input_format)
            if img_data is None:
                return FaceValidationResult(message=self.message.IS_VALID_ERROR)
            return self.face_factor.is_valid(image_data=img_data, config_object=config)
        except Exception as e:
            print("Oops: {}\nTrace: {}".format(e, traceback.format_exc()))
            print("Issue Tracker:: \nhttps://github.com/prividentity/cryptonets-python-sdk/issues")
            return FaceValidationResult(message=self.message.IS_VALID_ERROR)

    def estimate_age(self, image_path: str = None, image_data: np.array = None,
                     config: ConfigObject = None) -> FaceValidationResult:
        """Check if the image is valid and returns the age of the image

        Parameters
        ----------
        image_path
            Directory path to the image file

        config (Optional)
            Additional configuration parameters for the operation

        image_data (Optional)
            Image data in numpy RGB format

        Returns
        -------
        FaceValidationResult
            error: int [0 if successful -1 if any error]

            message: str [Message from the operation]

            face_objects: List[FaceObjectResult]

        """
        try:
            if config is not None and PARAMETERS.INPUT_IMAGE_FORMAT in config.config_param:
                input_format = config.config_param[PARAMETERS.INPUT_IMAGE_FORMAT]
            elif self.config is not None and PARAMETERS.INPUT_IMAGE_FORMAT in self.config.config_param:
                input_format = self.config.config_param[PARAMETERS.INPUT_IMAGE_FORMAT]
            else:
                input_format = "rgb"
            if (image_path is not None and image_data is not None) or (image_path is None and image_data is None):
                return FaceValidationResult(message="Specify either image_path or image_data")
            img_data = None
            if image_data is not None:
                if not isinstance(image_data, np.ndarray):
                    return FaceValidationResult(message="Required numpy array in RGB/RGBA/BGR format")
                img_data = image_data
            if image_path is not None and len(image_path) > 0:
                if not os.path.exists(image_path):
                    return FaceValidationResult(message=self.message.get_message(101))
                img_data = image_path_to_array(image_path, input_format=input_format)
            if img_data is None:
                return FaceValidationResult(message=self.message.AGE_ESTIMATE_ERROR)
            return self.face_factor.estimate_age(image_data=img_data, config_object=config)
        except Exception as e:
            print("Oops: {}\nTrace: {}".format(e, traceback.format_exc()))
            print("Issue Tracker:: \nhttps://github.com/prividentity/cryptonets-python-sdk/issues")
            return FaceValidationResult(message=self.message.AGE_ESTIMATE_ERROR)

    def enroll(self, image_path: str = None, image_data: np.array = None,
               config: ConfigObject = None) -> FaceEnrollPredictResult:
        """Enrolls the image in the face recognition server

        Parameters
        ----------
        image_path
            Directory path to the image file

        config (Optional)
            Additional configuration parameters for the operation

        image_data (Optional)
            Image data in numpy RGB format

        Returns
        -------
        FaceEnrollPredictResult
            status: int [0 if successful -1 if unsuccessful]

            message: str [Message from the operation]

            enroll_level: str

            guid: str

            uuid: str

            token: str
        """

        try:
            if config is not None and PARAMETERS.INPUT_IMAGE_FORMAT in config.config_param:
                input_format = config.config_param[PARAMETERS.INPUT_IMAGE_FORMAT]
            elif self.config is not None and PARAMETERS.INPUT_IMAGE_FORMAT in self.config.config_param:
                input_format = self.config.config_param[PARAMETERS.INPUT_IMAGE_FORMAT]
            else:
                input_format = "rgb"
            if (image_path is not None and image_data is not None) or (image_path is None and image_data is None):
                return FaceEnrollPredictResult(message="Specify either image_path or image_data")
            img_data = None
            if image_data is not None:
                if not isinstance(image_data, np.ndarray):
                    return FaceEnrollPredictResult(message="Required numpy array in RGB/RGBA/BGR format")
                img_data = image_data
            if image_path is not None and len(image_path) > 0:
                if not os.path.exists(image_path):
                    return FaceEnrollPredictResult(message=self.message.get_message(101))
                img_data = image_path_to_array(image_path, input_format=input_format)
            if img_data is None:
                return FaceEnrollPredictResult(message=self.message.EXCEPTION_ERROR_ENROLL)
            return self.face_factor.enroll(image_data=img_data, config_object=config)
        except Exception as e:
            print("Oops: {}\nTrace: {}".format(e, traceback.format_exc()))
            print("Issue Tracker:: \nhttps://github.com/prividentity/cryptonets-python-sdk/issues")
            return FaceEnrollPredictResult(message=self.message.EXCEPTION_ERROR_ENROLL)

    def get_iso_face(self, image_path: str = None, image_data: np.array = None,
                     config: ConfigObject = None) -> ISOFaceResult:
        """Takes the face image and gives back the image in ISO Spec format

        Parameters
        ----------
        image_path
            Directory path to the image file

        config (Optional)
            Additional configuration parameters for the operation

        image_data (Optional)
            Image data in numpy RGB format

        Returns
        -------
        ISOFaceResult
            status: int [0 if successful -1 if unsuccessful]

            message: str [Message from the operation]

            image: PIL.Image

            confidence: float

            iso_image_width: str

            iso_image_height: str

            iso_image_channels: str

        """
        try:
            if config is not None and PARAMETERS.INPUT_IMAGE_FORMAT in config.config_param:
                input_format = config.config_param[PARAMETERS.INPUT_IMAGE_FORMAT]
            elif self.config is not None and PARAMETERS.INPUT_IMAGE_FORMAT in self.config.config_param:
                input_format = self.config.config_param[PARAMETERS.INPUT_IMAGE_FORMAT]
            else:
                input_format = "rgb"
            if (image_path is not None and image_data is not None) or (image_path is None and image_data is None):
                return ISOFaceResult(message="Specify either image_path or image_data")
            img_data = None
            if image_data is not None:
                if not isinstance(image_data, np.ndarray):
                    return ISOFaceResult(message="Required numpy array in RGB/RGBA/BGR format")
                img_data = image_data
            if image_path is not None and len(image_path) > 0:
                if not os.path.exists(image_path):
                    return ISOFaceResult(message=self.message.get_message(101))
                img_data = image_path_to_array(image_path, input_format=input_format)
            if img_data is None:
                return ISOFaceResult(message=self.message.EXCEPTION_ERROR_GET_ISO_FACE)
            return self.face_factor.get_iso_face(image_data=img_data, config_object=config)
        except Exception as e:
            print("Oops: {}\nTrace: {}".format(e, traceback.format_exc()))
            print("Issue Tracker:: \nhttps://github.com/prividentity/cryptonets-python-sdk/issues")
            return ISOFaceResult(message=self.message.EXCEPTION_ERROR_GET_ISO_FACE)

    def predict(self, image_path: str = None, image_data: np.array = None,
                config: ConfigObject = None) -> FaceEnrollPredictResult:
        """Predicts the image in the face recognition server

        Parameters
        ----------
        image_path
            Directory path to the image file

        config (Optional)
            Additional configuration parameters for the operation

        image_data (Optional)
            Image data in numpy RGB format

        Returns
        -------
        FaceEnrollPredictResult
            status: int [0 if successful -1 if unsuccessful]

            message: str [Message from the operation]

            enroll_level: str

            guid: str

            uuid: str

            token: str

        """
        try:
            if config is not None and PARAMETERS.INPUT_IMAGE_FORMAT in config.config_param:
                input_format = config.config_param[PARAMETERS.INPUT_IMAGE_FORMAT]
            elif self.config is not None and PARAMETERS.INPUT_IMAGE_FORMAT in self.config.config_param:
                input_format = self.config.config_param[PARAMETERS.INPUT_IMAGE_FORMAT]
            else:
                input_format = "rgb"
            if (image_path is not None and image_data is not None) or (image_path is None and image_data is None):
                return FaceEnrollPredictResult(message="Specify either image_path or image_data")
            img_data = None
            if image_data is not None:
                if not isinstance(image_data, np.ndarray):
                    return FaceEnrollPredictResult(message="Required numpy array in RGB/RGBA/BGR format")
                img_data = image_data
            if image_path is not None and len(image_path) > 0:
                if not os.path.exists(image_path):
                    return FaceEnrollPredictResult(message=self.message.get_message(101))
                img_data = image_path_to_array(image_path, input_format=input_format)
            if img_data is None:
                return FaceEnrollPredictResult(message=self.message.EXCEPTION_ERROR_PREDICT)
            return self.face_factor.predict(image_data=img_data, config_object=config)
        except Exception as e:
            print("Oops: {}\nTrace: {}".format(e, traceback.format_exc()))
            print("Issue Tracker:: \nhttps://github.com/prividentity/cryptonets-python-sdk/issues")
            return FaceEnrollPredictResult(message=self.message.EXCEPTION_ERROR_PREDICT)

    def delete(self, uuid: str) -> FaceDeleteResult:
        """Deletes the enrollment from the face recognition server

        Parameters
        ----------
        uuid
            UUID of the enrolled image

        Returns
        -------
        FaceDeleteResult
            status: int [0 if successful -1 if unsuccessful]

            message: str [Message from the operation]
        """
        try:
            if uuid is None:
                return FaceDeleteResult(message="Missing UUID")
            return self.face_factor.delete(uuid)
        except Exception as e:
            print("Oops: {}\nTrace: {}".format(e, traceback.format_exc()))
            print("Issue Tracker:: \nhttps://github.com/prividentity/cryptonets-python-sdk/issues")
            return FaceDeleteResult(message=self.message.EXCEPTION_ERROR_DELETE)

    def compare(self, image_path_1: str = None, image_path_2: str = None, image_data_1: np.array = None,
                image_data_2: np.array = None, config: ConfigObject = None) -> FaceCompareResult:
        """Check if the images are of same person or not

        Parameters
        ----------
        image_path_1
            Directory path to the first image file

        image_path_2
            Directory path to the second image file

        config (Optional)
            Additional configuration parameters for the operation

        image_data_1 (Optional)
            First Image data in numpy RGB format

        image_data_2 (Optional)
            Second Image data in numpy RGB format

        Returns
        -------
        FaceCompareResult
            status: int [0 if same, 1 if different, -1 if unsuccessful]

            message: str [Message from the operation]

            result: str

            distance_min: str

            distance_mean: str

            distance_max: str

            first_validation_result: str

            second_validation_result: str

        """

        try:
            if config is not None and PARAMETERS.INPUT_IMAGE_FORMAT in config.config_param:
                input_format = config.config_param[PARAMETERS.INPUT_IMAGE_FORMAT]
            elif self.config is not None and PARAMETERS.INPUT_IMAGE_FORMAT in self.config.config_param:
                input_format = self.config.config_param[PARAMETERS.INPUT_IMAGE_FORMAT]
            else:
                input_format = "rgb"
            if (image_path_1 is not None and image_path_2 is not None and
                image_data_1 is not None and image_data_2 is not None) or (
                    image_path_1 is None and image_path_2 is None and image_data_1 is None and image_data_2 is None):
                return FaceCompareResult(message="Specify either image_path or image_data")
            img_data_1, img_data_2 = None, None
            if image_data_1 is not None and image_data_2 is not None:
                if not isinstance(image_data_1, np.ndarray) or not isinstance(image_data_2, np.ndarray):
                    return FaceCompareResult(message="Required numpy array in RGB/RGBA/BGR format")
                img_data_1 = image_data_1
                img_data_2 = image_data_2
            if (image_path_1 is not None and len(image_path_1) > 0) \
                    or (image_path_2 is not None and len(image_path_2) > 0):
                if not os.path.exists(image_path_1) or not os.path.exists(image_path_2):
                    return FaceCompareResult(message=self.message.get_message(101))
                img_data_1 = image_path_to_array(image_path_1, input_format=input_format)
                img_data_2 = image_path_to_array(image_path_2, input_format=input_format)
                if img_data_1 is None or img_data_2 is None:
                    return FaceCompareResult(message=self.message.EXCEPTION_ERROR_COMPARE)
            return self.face_factor.compare(image_data_1=img_data_1, image_data_2=img_data_2,
                                            config_object=config)
        except Exception as e:
            print("Oops: {}\nTrace: {}".format(e, traceback.format_exc()))
            print("Issue Tracker:: \nhttps://github.com/prividentity/cryptonets-python-sdk/issues")
            return FaceCompareResult(message=self.message.EXCEPTION_ERROR_COMPARE)

    @property
    def api_key(self) -> str:
        return self._api_key

    @property
    def server_url(self) -> str:
        return self._server_url

    @property
    def local_storage_path(self) -> str:
        return self._local_storage_path

    @property
    def logging_level(self) -> LoggingLevel:
        return self._logging_level

    @property
    def config(self) -> ConfigObject:
        return self._config_object

    @api_key.setter
    def api_key(self, value):
        self._api_key = value

    @server_url.setter
    def server_url(self, value):
        self._server_url = value

    @local_storage_path.setter
    def local_storage_path(self, value):
        self._local_storage_path = value

    @logging_level.setter
    def logging_level(self, value):
        self._logging_level = value

    @config.setter
    def config(self, value):
        self._config_object = value


if __name__ == "__main__":
    pass



class ISOFaceResult:
    def __init__(self, iso_image_width=None, iso_image_height=None, iso_image_channels=None, confidence=None,
                 image=None, status=-1, message=""):
        self._iso_image_width = iso_image_width
        self._iso_image_height = iso_image_height
        self._iso_image_channels = iso_image_channels
        self._confidence = confidence
        self._image = image
        self._status = status
        self._message = message

    @property
    def iso_image_width(self) -> int:
        """
        Returns the width of the Image object
        """
        return self._iso_image_width

    @property
    def iso_image_height(self) -> str:
        """
        Returns the height of the Image object
        """
        return self._iso_image_height

    @property
    def iso_image_channels(self) -> str:
        """
        Returns the number of channels present in the image
        """
        return self._iso_image_channels

    @property
    def confidence(self) -> str:
        """
        Returns the confidence score for the converted ISO image
        """
        return self._confidence

    @property
    def image(self) -> str:
        """
        Returns the PIL Image object face ISO Specification converted
        """
        return self._image

    @property
    def status(self) -> int:
        """
        Returns the status of the operation
        """
        return self._status

    @property
    def message(self) -> str:
        """
        Returns the message of the operation
        """
        return self._message

    @iso_image_width.setter
    def iso_image_width(self, value):
        self._iso_image_width = value

    @iso_image_width.setter
    def iso_image_width(self, value):
        self._iso_image_width = value

    @iso_image_height.setter
    def iso_image_height(self, value):
        self._iso_image_height = value

    @iso_image_channels.setter
    def iso_image_channels(self, value):
        self._iso_image_channels = value

    @status.setter
    def status(self, value):
        self._status = value

    @message.setter
    def message(self, value):
        self._message = value



#  PRIVATE IDENTITY LLC - PROPRIETARY AND CONFIDENTIAL (c) COPYRIGHT 2018 - 2020 PRIVATE IDENTITY LLC
#  All Rights Reserved.
#  NOTICE:  All information contained herein is, and remains the
#  property of PRIVATE IDENTITY LLC and its suppliers, if any.  The intellectual and technical concepts contained
#  herein are proprietary to PRIVATE IDENTITY LLC and its suppliers and may be covered by U.S. and Foreign Patents,
#  patents in process, and are protected by trade secret or copyright law. Dissemination of this information or
#  reproduction of this material is strictly forbidden unless prior written permission is obtained from PRIVATE
#  IDENTITY LLC. Any software that is made available to download from Private Identity LLC ("Software") is the
#  copyrighted work of Private Identity LLC and/or its suppliers. Use of the Software is governed by the terms of the
#  end user license agreement, if any, which accompanies or is included with the Software ("License Agreement"). An
#  end user must not install any Software that is accompanied by or includes a License Agreement, unless he or she
#  first agrees to the License Agreement terms. RESTRICTED RIGHTS LEGEND. Any Software which is downloaded from
#  PRIVATE IDENTITY LLC for or on behalf of the United States of America, its agencies and/or instrumentalities ("U.S.
#  Government"), is provided with Restricted Rights. Use, duplication, or disclosure by the U.S. Government is subject
#  to restrictions as set forth in subparagraph (c)(1)(ii) of the Rights in Technical Data and Computer Software
#  clause at DFARS 252.227-7013 or subparagraphs (c)(1) and (2) of the Commercial Computer Software - Restricted
#  Rights at 48 CFR 52.227-19, as applicable.
#  Manufacturer is PRIVATE IDENTITY LLC , 13331 Signal Tree, Potomac, MD  20854  USA.

import functools
import warnings


class Singleton(type):
    _instances = {}

    def __call__(cls, *args, **kwargs):
        if cls not in cls._instances:
            cls._instances[cls] = super(Singleton, cls).__call__(*args, **kwargs)
        return cls._instances[cls]


def deprecated(func):
    """This is a decorator which can be used to mark functions
    as deprecated. It will result in a warning being emitted
    when the function is used."""

    @functools.wraps(func)
    def new_func(*args, **kwargs):
        warnings.simplefilter('always', DeprecationWarning)  # turn off filter
        warnings.warn("Call to deprecated function {}.".format(func.__name__),
                      category=DeprecationWarning,
                      stacklevel=2)
        warnings.simplefilter('default', DeprecationWarning)  # reset filter
        return func(*args, **kwargs)

    return new_func


def log_error(logger):
    def decorated(f):
        @functools.wraps(f)
        def wrapped(*args, **kwargs):
            try:
                return f(*args, **kwargs)
            except Exception as e:
                if logger:
                    logger.exception(e)
                raise e

        return wrapped

    return decorated




import platform

from ..settings.loggingLevel import LoggingLevel
from ..settings.supportedPlatforms import SupportedPlatforms
import sys

class ModuleSettings:
    def __init__(self, api_key=None, server_url=None, local_storage_path="privateid",
                 logging_level=LoggingLevel.off.value):

        try:
            if not api_key or not server_url or not local_storage_path or logging_level is None:
                raise ValueError("Wrong parameters while initializing the SDK")
            if platform.system() not in SupportedPlatforms.supportedOS.value:
                raise OSError("Invalid OS")

            self._api_key = bytes(api_key, 'utf-8')
            self._local_storage_path = bytes(local_storage_path, 'utf-8')
            self._logging_level = logging_level
            if server_url[-1] == "/":
                server_url=server_url[:-1]
            self._server_url = bytes(server_url, 'utf-8')
        except ValueError as exp:
            print("Error", exp)
            sys.exit(1)
        except OSError as exp:
            print("Error", exp)
            sys.exit(1)
    @property
    def api_key(self):
        return self._api_key

    @property
    def server_url(self):
        return self._server_url

    @property
    def local_storage_path(self):
        return self._local_storage_path

    @property
    def logging_level(self):
        return self._logging_level



from typing import List

from ..result_objects.faceObjectResult import FaceObjectResult


class FaceValidationResult:
    def __init__(self, error=-1, message=""):
        """Face validation result class for handling the detected faces
        """
        self._error = error
        self._message = message
        self._face_objects = []

    @property
    def error(self) -> int:
        """
        Returns the status of the operation
        """
        return self._error

    @property
    def message(self) -> str:
        """
        Returns the message of the operation
        """
        return self._message

    @property
    def face_objects(self) -> List[FaceObjectResult]:
        """
        Returns the list of Face Objects
        """
        return self._face_objects

    def append_face_objects(self, return_code=-100, message="", age=None, top_left_coordinate=None,
                            bottom_right_coordinate=None):
        self._face_objects.append(
            FaceObjectResult(return_code=return_code, message=message, age=age, top_left_coordinate=top_left_coordinate,
                             bottom_right_coordinate=bottom_right_coordinate))

    @message.setter
    def message(self, value):
        self._message = value

    @face_objects.setter
    def face_objects(self, value):
        self._face_objects = value



class FaceIsValidDeprecatedResult:
    def __init__(self, result=None, age_factor=None, output_image_data=None, status=-1, message=""):
        self._status = status
        self._message = message
        self._result = result
        self._age_factor = age_factor
        self._output_image_data = output_image_data

    @property
    def status(self) -> int:
        """
        Returns the status of the operation

        0 - If successfully obtained result from server

        -1 - In case of error

        """
        return self._status

    @property
    def message(self) -> str:
        """
        Returns the message of the operation
        """
        return self._message

    @property
    def result(self) -> int:
        """
        Returns the result of the operation
        """
        return self._result

    @property
    def age_factor(self) -> float:
        """
        Returns the predicted age of the image

        Age value might be in the float format as it is returned from DNN model
        """
        return self._age_factor

    @property
    def output_image_data(self):
        """
        Returns the cropped image data from the image

        This is optional, configuration has to be implemented to extract data
        """
        return self._output_image_data

    @status.setter
    def status(self, value):
        self._status = value

    @message.setter
    def message(self, value):
        self._message = value

    @result.setter
    def result(self, value):
        self._result = value

    @age_factor.setter
    def age_factor(self, value):
        self._age_factor = value

    @output_image_data.setter
    def output_image_data(self, value):
        self._output_image_data = value



# -*- coding: utf-8 -*-
"""
 PRIVATE IDENTITY LLC - PROPRIETARY AND CONFIDENTIAL (c) COPYRIGHT 2018 - 2020 PRIVATE IDENTITY LLC
 All Rights Reserved.
 NOTICE:  All information contained herein is, and remains the
 property of PRIVATE IDENTITY LLC and its suppliers, if any.  The intellectual and technical concepts contained
 herein are proprietary to PRIVATE IDENTITY LLC and its suppliers and may be covered by U.S. and Foreign Patents,
 patents in process, and are protected by trade secret or copyright law. Dissemination of this information or
 reproduction of this material is strictly forbidden unless prior written permission is obtained from PRIVATE
 IDENTITY LLC. Any software that is made available to download from Private Identity LLC ("Software") is the
 copyrighted work of Private Identity LLC and/or its suppliers. Use of the Software is governed by the terms of the
 end user license agreement, if any, which accompanies or is included with the Software ("License Agreement"). An
 end user must not install any Software that is accompanied by or includes a License Agreement, unless he or she
 first agrees to the License Agreement terms. RESTRICTED RIGHTS LEGEND. Any Software which is downloaded from
 PRIVATE IDENTITY LLC for or on behalf of the United States of America, its agencies and/or instrumentalities ("U.S.
 Government"), is provided with Restricted Rights. Use, duplication, or disclosure by the U.S. Government is subject
 to restrictions as set forth in subparagraph (c)(1)(ii) of the Rights in Technical Data and Computer Software
 clause at DFARS 252.227-7013 or subparagraphs (c)(1) and (2) of the Commercial Computer Software - Restricted
 Rights at 48 CFR 52.227-19, as applicable.
 Manufacturer is PRIVATE IDENTITY LLC , 13331 Signal Tree, Potomac, MD  20854  USA.
"""

import os

from setuptools import setup, find_packages

"""
Setup module for cryptonets python sdk.

Created on 22/11/2022

@author: Private Identity
"""
NAME = "cryptonets_python_sdk"
DESCRIPTION = "Cryptonets SDK Library for Python"
AUTHOR = "Private Identity"
AUTHOR_EMAIL = "support@private.id"
URL = "https://private.id/"
VERSION = "1.0.15"
REQUIRES = [
    "numpy >= 1.23.0",
    "pillow >= 9.1.0"
]

LONG_DESCRIPTION = ''
if os.path.exists('./README.md'):
    with open("README.md", encoding='utf-8') as fp:
        LONG_DESCRIPTION = fp.read()

setup(
    name=NAME,
    version=VERSION,
    description=DESCRIPTION,
    long_description=LONG_DESCRIPTION,
    long_description_content_type='text/markdown',
    author=AUTHOR,
    author_email=AUTHOR_EMAIL,
    license="",
    url=URL,
    keywords=["privateid", "cryptonets", "face identification"],
    packages=find_packages(where='src', exclude=["tests*"]),
    include_package_data=True,
    platforms="any",
    install_requires=REQUIRES,
    python_requires=">=3.6",
    package_dir={'': 'src'},
    project_urls={
        "Bug Reports": 'https://github.com/prividentity/cryptonets-python-sdk/issues',
        "Source": 'https://github.com/prividentity/cryptonets-python-sdk',
        "Documentation": "https://docs.private.id/cryptonets-python-sdk/index.html",
        "Release Notes": "https://docs.private.id/cryptonets-python-sdk/changelog.html"
    },
    classifiers=[
        "Intended Audience :: Developers",
        "License :: Other/Proprietary License",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.6",
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        "Topic :: Software Development"
    ],
    py_modules=[NAME],
    package_data={
        "cryptonets_python_sdk": ['handler/lib/lib_fhe.so']
    }
)



class FaceEnrollPredictResult:
    def __init__(self, enroll_level=None, uuid=None, guid=None, token=None, status=-1, message=""):
        self._enroll_level = enroll_level
        self._uuid = uuid
        self._guid = guid
        self._token = token
        self._status = status
        self._message = message

    @property
    def enroll_level(self) -> int:
        """
        Returns the enroll_level of the user
        0 - one-factor authentication

        1 - two-factor authentication
        """
        return self._enroll_level

    @property
    def uuid(self) -> str:
        """
        Returns the UUID of the user

        Unique ID of length 20
        """
        return self._uuid

    @property
    def guid(self) -> str:
        """
        Returns the GUID of the user

        Unique ID of length 20
        """
        return self._guid

    @property
    def token(self) -> str:
        """
        Returns the token for verifying the operation

        Yet to be implemented for configuration
        """
        return self._token

    @property
    def status(self) -> int:
        """
        Returns the status of the operation

        0 - If successfully obtained result from server

        -1 - In case of error

        """
        return self._status

    @property
    def message(self) -> str:
        """
        Returns the message of the operation
        """
        return self._message

    @enroll_level.setter
    def enroll_level(self, value):
        self._enroll_level = value

    @guid.setter
    def guid(self, value):
        self._guid = value

    @uuid.setter
    def uuid(self, value):
        self._uuid = value

    @status.setter
    def status(self, value):
        self._status = value

    @token.setter
    def token(self, value):
        self._token = value

    @message.setter
    def message(self, value):
        self._message = value



class FaceIsValidResult:
    def __init__(self,status=-1, message=""):
        self._status = status
        self._message = message

    @property
    def status(self):
        """
        Returns the status of the operation
        """
        return self._status

    @property
    def message(self):
        """
        Returns the message of the operation
        """
        return self._message

    @status.setter
    def status(self, value):
        self._status = value

    @message.setter
    def message(self, value):
        self._message = value



class FaceCompareResult:
    def __init__(self, result=None, distance_min=None, distance_mean=None, distance_max=None,
                 second_validation_result=None,
                 first_validation_result=None, status=-1, message=""):
        """Result handler for compare
        """
        self._status = status
        self._result = result
        self._message = message
        self._distance_min = distance_min
        self._distance_mean = distance_mean
        self._distance_max = distance_max
        self._first_validation_result = first_validation_result
        self._second_validation_result = second_validation_result

    @property
    def status(self) -> int:
        """
        Returns the status of the operation

        0 - If successfully obtained result from server

        -1 - In case of error

        """
        return self._status

    @property
    def result(self) -> int:
        """
        Returns the result of the operation
        """
        return self._result

    @property
    def message(self) -> str:
        """
        Returns the message of the operation
        """
        return self._message

    @property
    def distance_min(self) -> float:
        """
        Returns the minimum distance

        Minimum distance refers to the thresholds used for comparing the images

        """
        return self._distance_min

    @property
    def distance_mean(self) -> float:
        """
        Returns the average distance

        Mean distance refers to the thresholds used for comparing the images

        """
        return self._distance_mean

    @property
    def distance_max(self) -> float:
        """
        Returns the maximum distance

        Maximum distance refers to the thresholds used for comparing the images

        """
        return self._distance_max

    @property
    def first_validation_result(self) -> int:
        """
        Returns the validation result of first image
        """
        return self._first_validation_result

    @property
    def second_validation_result(self) -> int:
        """
        Returns the validation result of second image
        """
        return self._second_validation_result

    @status.setter
    def status(self, value):
        self._status = value

    @result.setter
    def result(self, value):
        self._result = value

    @message.setter
    def message(self, value):
        self._message = value

    @distance_min.setter
    def distance_min(self, value):
        self._distance_min = value

    @distance_mean.setter
    def distance_mean(self, value):
        self._distance_mean = value

    @distance_max.setter
    def distance_max(self, value):
        self._distance_max = value

    @first_validation_result.setter
    def first_validation_result(self, value):
        self._first_validation_result = value

    @second_validation_result.setter
    def second_validation_result(self, value):
        self._second_validation_result = value
